/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComSlice.cpp
    \brief    slice header and SPS class
*/

#include "CommonDef.h"
#include "TComSlice.h"
#include "TComPic.h"

TComSlice::TComSlice()
{
  m_iPOC                = 0;
  m_eSliceType          = I_SLICE;
  m_iSliceQp            = 0;
  m_iSymbolMode         = 1;
  m_aiNumRefIdx[0]      = 0;
  m_aiNumRefIdx[1]      = 0;
  m_bLoopFilterDisable  = false;
  
  m_bDRBFlag            = true;
  m_eERBIndex           = ERB_NONE;
  
  m_iSliceQpDelta       = 0;
  
  m_iDepth              = 0;
  
  m_pcPic               = NULL;
  m_bRefenced           = false;
#ifdef ROUNDING_CONTROL_BIPRED
  m_bRounding           = false;
#endif
  m_uiColDir = 0;
  
  initEqualRef();
  m_bNoBackPredFlag = false;
#if MS_LCEC_LOOKUP_TABLE_EXCEPTION
  m_bRefIdxCombineCoding = false;
#endif
#if DCM_COMB_LIST 
  m_bRefPicListCombinationFlag = false;
  m_bRefPicListModificationFlagLC = false;
#endif
  m_uiSliceCurStartCUAddr        = 0;
  m_uiEntropySliceCurStartCUAddr = 0;
}

TComSlice::~TComSlice()
{
}


Void TComSlice::initSlice()
{
  m_aiNumRefIdx[0]      = 0;
  m_aiNumRefIdx[1]      = 0;
  
  m_bDRBFlag            = true;
  m_eERBIndex           = ERB_NONE;
  
  m_uiColDir = 0;
  
  initEqualRef();
  m_bNoBackPredFlag = false;
#if MS_LCEC_LOOKUP_TABLE_EXCEPTION
  m_bRefIdxCombineCoding = false;
#endif
#if DCM_COMB_LIST 
  m_bRefPicListCombinationFlag = false;
  m_bRefPicListModificationFlagLC = false;
#endif
}

Void  TComSlice::sortPicList        (TComList<TComPic*>& rcListPic)
{
  TComPic*    pcPicExtract;
  TComPic*    pcPicInsert;
  
  TComList<TComPic*>::iterator    iterPicExtract;
  TComList<TComPic*>::iterator    iterPicExtract_1;
  TComList<TComPic*>::iterator    iterPicInsert;
  
  for (Int i = 1; i < (Int)(rcListPic.size()); i++)
  {
    iterPicExtract = rcListPic.begin();
    for (Int j = 0; j < i; j++) iterPicExtract++;
    pcPicExtract = *(iterPicExtract);
    pcPicExtract->setCurrSliceIdx(0);
    
    iterPicInsert = rcListPic.begin();
    while (iterPicInsert != iterPicExtract)
    {
      pcPicInsert = *(iterPicInsert);
      pcPicInsert->setCurrSliceIdx(0);
      if (pcPicInsert->getPOC() >= pcPicExtract->getPOC())
      {
        break;
      }
      
      iterPicInsert++;
    }
    
    iterPicExtract_1 = iterPicExtract;    iterPicExtract_1++;
    
    //  swap iterPicExtract and iterPicInsert, iterPicExtract = curr. / iterPicInsert = insertion position
    rcListPic.insert (iterPicInsert, iterPicExtract, iterPicExtract_1);
    rcListPic.erase  (iterPicExtract);
  }
}

TComPic* TComSlice::xGetRefPic (TComList<TComPic*>& rcListPic,
                                Bool                bDRBFlag,
                                ERBIndex            eERBIndex,
                                UInt                uiPOCCurr,
                                RefPicList          eRefPicList,
                                UInt                uiNthRefPic)
{
  //  find current position
  TComList<TComPic*>::iterator  iterPic = rcListPic.begin();
  TComPic*                      pcRefPic   = NULL;
  
  TComPic*                      pcPic = *(iterPic);
  while ( (pcPic->getPOC() != (Int)uiPOCCurr) && (iterPic != rcListPic.end()) )
  {
    iterPic++;
    pcPic = *(iterPic);
  }
  assert (pcPic->getPOC() == (Int)uiPOCCurr);
  
  //  find n-th reference picture with bDRBFlag and eERBIndex
  UInt  uiCount = 0;
  
  if( eRefPicList == REF_PIC_LIST_0 )
  {
    while(1)
    {
      if (iterPic == rcListPic.begin())
        break;
      
      iterPic--;
      pcPic = *(iterPic);
      if( ( !pcPic->getReconMark()                        ) ||
          ( bDRBFlag  != pcPic->getSlice(0)->getDRBFlag()  ) ||
          ( eERBIndex != pcPic->getSlice(0)->getERBIndex() ) )
        continue;
      
      if( !pcPic->getSlice(0)->isReferenced() )
        continue;
      
      uiCount++;
      if (uiCount == uiNthRefPic)
      {
        pcRefPic = pcPic;
        return  pcRefPic;
      }
    }
    
    if ( !m_pcSPS->getUseLDC() )
    {
      
      iterPic = rcListPic.begin();
      pcPic = *(iterPic);
      while ( (pcPic->getPOC() != (Int)uiPOCCurr) && (iterPic != rcListPic.end()) )
      {
        iterPic++;
        pcPic = *(iterPic);
      }
      assert (pcPic->getPOC() == (Int)uiPOCCurr);
      
      while(1)
      {
        iterPic++;
        if (iterPic == rcListPic.end())
          break;
        
        pcPic = *(iterPic);
        if( ( !pcPic->getReconMark()                        ) ||
          ( bDRBFlag  != pcPic->getSlice(0)->getDRBFlag()  ) ||
          ( eERBIndex != pcPic->getSlice(0)->getERBIndex() ) )
          continue;
        
      if( !pcPic->getSlice(0)->isReferenced() )
          continue;
        
        uiCount++;
        if (uiCount == uiNthRefPic)
        {
          pcRefPic = pcPic;
          return  pcRefPic;
        }
      }
    }
  }
  else
  {
    while(1)
    {
      iterPic++;
      if (iterPic == rcListPic.end())
        break;
      
      pcPic = *(iterPic);
      if( ( !pcPic->getReconMark()                        ) ||
          ( bDRBFlag  != pcPic->getSlice(0)->getDRBFlag()  ) ||
          ( eERBIndex != pcPic->getSlice(0)->getERBIndex() ) )
        continue;
      
      if( !pcPic->getSlice(0)->isReferenced() )
        continue;
      
      uiCount++;
      if (uiCount == uiNthRefPic)
      {
        pcRefPic = pcPic;
        return  pcRefPic;
      }
    }
    
    iterPic = rcListPic.begin();
    pcPic = *(iterPic);
    while ( (pcPic->getPOC() != (Int)uiPOCCurr) && (iterPic != rcListPic.end()) )
    {
      iterPic++;
      pcPic = *(iterPic);
    }
    assert (pcPic->getPOC() == (Int)uiPOCCurr);
    
    while(1)
    {
      if (iterPic == rcListPic.begin())
        break;
      
      iterPic--;
      pcPic = *(iterPic);
      if( ( !pcPic->getReconMark()                        ) ||
          ( bDRBFlag  != pcPic->getSlice(0)->getDRBFlag()  ) ||
          ( eERBIndex != pcPic->getSlice(0)->getERBIndex() ) )
        continue;
      
      if( !pcPic->getSlice(0)->isReferenced() )
        continue;
      
      uiCount++;
      if (uiCount == uiNthRefPic)
      {
        pcRefPic = pcPic;
        return  pcRefPic;
      }
    }
  }
  
  return  pcRefPic;
}

Void TComSlice::setRefPOCList       ()
{
  for (Int iDir = 0; iDir < 2; iDir++)
  {
    for (Int iNumRefIdx = 0; iNumRefIdx < m_aiNumRefIdx[iDir]; iNumRefIdx++)
    {
      m_aiRefPOCList[iDir][iNumRefIdx] = m_apcRefPicList[iDir][iNumRefIdx]->getPOC();
    }
  }

}

#if DCM_COMB_LIST 
Void TComSlice::generateCombinedList()
{
  if(m_aiNumRefIdx[REF_PIC_LIST_C] > 0)
  {
    m_aiNumRefIdx[REF_PIC_LIST_C]=0;
    for(Int iNumCount = 0; iNumCount < MAX_NUM_REF_LC; iNumCount++)
    {
      m_iRefIdxOfLC[REF_PIC_LIST_0][iNumCount]=-1;
      m_iRefIdxOfLC[REF_PIC_LIST_1][iNumCount]=-1;
      m_eListIdFromIdxOfLC[iNumCount]=0;
      m_iRefIdxFromIdxOfLC[iNumCount]=0;
      m_iRefIdxOfL0FromRefIdxOfL1[iNumCount] = -1;
      m_iRefIdxOfL1FromRefIdxOfL0[iNumCount] = -1;
    }

    for (Int iNumRefIdx = 0; iNumRefIdx < MAX_NUM_REF; iNumRefIdx++)
    {
      if(iNumRefIdx < m_aiNumRefIdx[REF_PIC_LIST_0]){
        Bool bTempRefIdxInL2 = true;
        for ( Int iRefIdxLC = 0; iRefIdxLC < m_aiNumRefIdx[REF_PIC_LIST_C]; iRefIdxLC++ )
        {
          if ( m_apcRefPicList[REF_PIC_LIST_0][iNumRefIdx]->getPOC() == m_apcRefPicList[m_eListIdFromIdxOfLC[iRefIdxLC]][m_iRefIdxFromIdxOfLC[iRefIdxLC]]->getPOC() )
          {
            m_iRefIdxOfL1FromRefIdxOfL0[iNumRefIdx] = m_iRefIdxFromIdxOfLC[iRefIdxLC];
            m_iRefIdxOfL0FromRefIdxOfL1[m_iRefIdxFromIdxOfLC[iRefIdxLC]] = iNumRefIdx;
            bTempRefIdxInL2 = false;
            assert(m_eListIdFromIdxOfLC[iRefIdxLC]==REF_PIC_LIST_1);
            break;
          }
        }

        if(bTempRefIdxInL2 == true)
        { 
          m_eListIdFromIdxOfLC[m_aiNumRefIdx[REF_PIC_LIST_C]] = REF_PIC_LIST_0;
          m_iRefIdxFromIdxOfLC[m_aiNumRefIdx[REF_PIC_LIST_C]] = iNumRefIdx;
          m_iRefIdxOfLC[REF_PIC_LIST_0][iNumRefIdx] = m_aiNumRefIdx[REF_PIC_LIST_C]++;
        }
      }

      if(iNumRefIdx < m_aiNumRefIdx[REF_PIC_LIST_1]){
        Bool bTempRefIdxInL2 = true;
        for ( Int iRefIdxLC = 0; iRefIdxLC < m_aiNumRefIdx[REF_PIC_LIST_C]; iRefIdxLC++ )
        {
          if ( m_apcRefPicList[REF_PIC_LIST_1][iNumRefIdx]->getPOC() == m_apcRefPicList[m_eListIdFromIdxOfLC[iRefIdxLC]][m_iRefIdxFromIdxOfLC[iRefIdxLC]]->getPOC() )
          {
            m_iRefIdxOfL0FromRefIdxOfL1[iNumRefIdx] = m_iRefIdxFromIdxOfLC[iRefIdxLC];
            m_iRefIdxOfL1FromRefIdxOfL0[m_iRefIdxFromIdxOfLC[iRefIdxLC]] = iNumRefIdx;
            bTempRefIdxInL2 = false;
            assert(m_eListIdFromIdxOfLC[iRefIdxLC]==REF_PIC_LIST_0);
            break;
          }
        }
        if(bTempRefIdxInL2 == true)
        {
          m_eListIdFromIdxOfLC[m_aiNumRefIdx[REF_PIC_LIST_C]] = REF_PIC_LIST_1;
          m_iRefIdxFromIdxOfLC[m_aiNumRefIdx[REF_PIC_LIST_C]] = iNumRefIdx;
          m_iRefIdxOfLC[REF_PIC_LIST_1][iNumRefIdx] = m_aiNumRefIdx[REF_PIC_LIST_C]++;
        }
      }
    }
  }
}
#endif

Void TComSlice::setRefPicList       ( TComList<TComPic*>& rcListPic )
{
  if (m_eSliceType == I_SLICE)
  {
    ::memset( m_apcRefPicList, 0, sizeof (m_apcRefPicList));
    ::memset( m_aiNumRefIdx,   0, sizeof ( m_aiNumRefIdx ));
    
    return;
  }
  
  m_aiNumRefIdx[0] = Min ( m_aiNumRefIdx[0], (Int)(rcListPic.size())-1 );
  m_aiNumRefIdx[1] = Min ( m_aiNumRefIdx[1], (Int)(rcListPic.size())-1 );
  
  sortPicList(rcListPic);
  
  TComPic*  pcRefPic;
  for (Int n = 0; n < 2; n++)
  {
    RefPicList  eRefPicList = (RefPicList)n;
    
    UInt  uiOrderDRB  = 1;
    UInt  uiOrderERB  = 1;
    Int   iRefIdx     = 0;
    UInt  uiActualListSize = 0;
    
    if ( m_eSliceType == P_SLICE && eRefPicList == REF_PIC_LIST_1 )
    {
      m_aiNumRefIdx[eRefPicList] = 0;
      ::memset( m_apcRefPicList[eRefPicList], 0, sizeof(m_apcRefPicList[eRefPicList]));
      break;
    }
    
    //  First DRB
    pcRefPic = xGetRefPic(rcListPic, true, ERB_NONE, m_iPOC, eRefPicList, uiOrderDRB);
    if (pcRefPic != NULL)
    {
      m_apcRefPicList[eRefPicList][iRefIdx] = pcRefPic;
      
      pcRefPic->getPicYuvRec()->extendPicBorder();
      
      iRefIdx++;
      uiOrderDRB++;
      uiActualListSize++;
    }
    
    if ( (Int)uiActualListSize >= m_aiNumRefIdx[eRefPicList] )
    {
      m_aiNumRefIdx[eRefPicList] = uiActualListSize;
      continue;
    }
    
    // Long term reference
    // Should be enabled to support long term refernce
    //*
    //  First ERB
    pcRefPic = xGetRefPic(rcListPic, false, ERB_LTR, m_iPOC, eRefPicList, uiOrderERB);
    if (pcRefPic != NULL)
    {
      Bool  bChangeDrbErb = false;
      if      (iRefIdx > 0 && eRefPicList == REF_PIC_LIST_0 && pcRefPic->getPOC() > m_apcRefPicList[eRefPicList][iRefIdx-1]->getPOC())
      {
        bChangeDrbErb = true;
      }
      else if (iRefIdx > 0 && eRefPicList == REF_PIC_LIST_1 && pcRefPic->getPOC() < m_apcRefPicList[eRefPicList][iRefIdx-1]->getPOC())
      {
        bChangeDrbErb = true;
      }
      
      if ( bChangeDrbErb)
      {
        m_apcRefPicList[eRefPicList][iRefIdx]   = m_apcRefPicList[eRefPicList][iRefIdx-1];
        m_apcRefPicList[eRefPicList][iRefIdx-1] = pcRefPic;
      }
      else
      {
        m_apcRefPicList[eRefPicList][iRefIdx] = pcRefPic;
      }
      
      pcRefPic->getPicYuvRec()->extendPicBorder();
      
      iRefIdx++;
      uiOrderERB++;
      uiActualListSize++;
    }
    //*/
    
    // Short term reference
    //  Residue DRB
    UInt  uiBreakCount = 17 - iRefIdx;
    while (1)
    {
      uiBreakCount--;
      if ( (Int)uiActualListSize >= m_aiNumRefIdx[eRefPicList] || uiBreakCount == 0 )
      {
        break;
      }
      
      pcRefPic = xGetRefPic(rcListPic, true, ERB_NONE, m_iPOC, eRefPicList, uiOrderDRB);
      if (pcRefPic != NULL)
      {
        uiOrderDRB++;
      }
      
      if (pcRefPic != NULL)
      {
        m_apcRefPicList[eRefPicList][iRefIdx] = pcRefPic;
        
        pcRefPic->getPicYuvRec()->extendPicBorder();
        
        iRefIdx++;
        uiActualListSize++;
      }
    }
    
    m_aiNumRefIdx[eRefPicList] = uiActualListSize;
  }
}

Void TComSlice::initEqualRef()
{
  for (Int iDir = 0; iDir < 2; iDir++)
  {
    for (Int iRefIdx1 = 0; iRefIdx1 < MAX_NUM_REF; iRefIdx1++)
    {
      for (Int iRefIdx2 = iRefIdx1; iRefIdx2 < MAX_NUM_REF; iRefIdx2++)
      {
        m_abEqualRef[iDir][iRefIdx1][iRefIdx2] = m_abEqualRef[iDir][iRefIdx2][iRefIdx1] = (iRefIdx1 == iRefIdx2? true : false);
      }
    }
  }
}

#if DCM_DECODING_REFRESH
/** Function for marking the reference pictures when an IDR and CDR is encountered.
 * \param uiPOCCDR POC of the CDR picture
 * \param bRefreshPending flag indicating if a deferred decoding refresh is pending
 * \param rcListPic reference to the reference picture list
 * This function marks the reference pictures as "unused for reference" in the following conditions.
 * If the nal_unit_type is IDR all pictures in the reference picture list  
 * is marked as "unused for reference" 
 * Otherwise do for the CDR case (non CDR case has no effect since both if conditions below will not be true)
 *    If the bRefreshPending flag is true (a deferred decoding refresh is pending) and the current 
 *    temporal reference is greater than the temporal reference of the latest CDR picture (uiPOCCDR), 
 *    mark all reference pictures except the latest CDR picture as "unused for reference" and set 
 *    the bRefreshPending flag to false.
 *    If the nal_unit_type is CDR, set the bRefreshPending flag to true and iPOCCDR to the temporal 
 *    reference of the current picture.
 * Note that the current picture is already placed in the reference list and its marking is not changed.
 * If the current picture has a nal_ref_idc that is not 0, it will remain marked as "used for reference".
 */
Void TComSlice::decodingRefreshMarking(UInt& uiPOCCDR, Bool& bRefreshPending, TComList<TComPic*>& rcListPic)
{
  TComPic*                 rpcPic;
  UInt uiPOCCurr = getPOC(); 

  if (getNalUnitType() == NAL_UNIT_CODED_SLICE_IDR)  // IDR
  {
    // mark all pictures as not used for reference
    TComList<TComPic*>::iterator        iterPic       = rcListPic.begin();
    while (iterPic != rcListPic.end())
    {
      rpcPic = *(iterPic);
      rpcPic->setCurrSliceIdx(0);
      if (rpcPic->getPOC() != uiPOCCurr) rpcPic->getSlice(0)->setReferenced(false);
      iterPic++;
    }
  }
  else // CDR or No DR
  {
    if (bRefreshPending==true && uiPOCCurr > uiPOCCDR) // CDR reference marking pending 
    {
      TComList<TComPic*>::iterator        iterPic       = rcListPic.begin();
      while (iterPic != rcListPic.end())
      {
        rpcPic = *(iterPic);
        if (rpcPic->getPOC() != uiPOCCurr && rpcPic->getPOC() != uiPOCCDR) rpcPic->getSlice(0)->setReferenced(false);
        iterPic++;
      }
      bRefreshPending = false; 
    }
    if (getNalUnitType() == NAL_UNIT_CODED_SLICE_CDR) // CDR picture found
    {
      bRefreshPending = true; 
      uiPOCCDR = uiPOCCurr;
    }
  }
}
#endif

Void TComSlice::copySliceInfo(TComSlice *pSrc)
{
  assert( pSrc != NULL );

  Int i, j, k;

  m_iPOC                 = pSrc->m_iPOC;
#if DCM_DECODING_REFRESH
  m_eNalUnitType         = pSrc->m_eNalUnitType;
#endif  
  m_eSliceType           = pSrc->m_eSliceType;
  m_iSliceQp             = pSrc->m_iSliceQp;
  m_iSymbolMode          = pSrc->m_iSymbolMode;
  m_bLoopFilterDisable   = pSrc->m_bLoopFilterDisable;
  m_bDRBFlag             = pSrc->m_bDRBFlag;
  m_eERBIndex            = pSrc->m_eERBIndex;
  
#if DCM_COMB_LIST  
  for (i = 0; i < 3; i++)
  {
    m_aiNumRefIdx[i]     = pSrc->m_aiNumRefIdx[i];
  }

  for (i = 0; i < 2; i++)
  {
    for (j = 0; j < MAX_NUM_REF_LC; j++)
    {
       m_iRefIdxOfLC[i][j]  = pSrc->m_iRefIdxOfLC[i][j];
    }
  }
  for (i = 0; i < MAX_NUM_REF_LC; i++)
  {
    m_eListIdFromIdxOfLC[i] = pSrc->m_eListIdFromIdxOfLC[i];
    m_iRefIdxFromIdxOfLC[i] = pSrc->m_iRefIdxFromIdxOfLC[i];
    m_iRefIdxOfL1FromRefIdxOfL0[i] = pSrc->m_iRefIdxOfL1FromRefIdxOfL0[i];
    m_iRefIdxOfL0FromRefIdxOfL1[i] = pSrc->m_iRefIdxOfL0FromRefIdxOfL1[i];
  }
  m_bRefPicListModificationFlagLC = pSrc->m_bRefPicListModificationFlagLC;
  m_bRefPicListCombinationFlag    = pSrc->m_bRefPicListCombinationFlag;
#else
  for (i = 0; i < 2; i++)
  {
    m_aiNumRefIdx[i]     = pSrc->m_aiNumRefIdx[i];
  }
#endif  

  m_iSliceQpDelta        = pSrc->m_iSliceQpDelta;
  for (i = 0; i < 2; i++)
  {
    for (j = 0; j < MAX_NUM_REF; j++)
    {
      m_apcRefPicList[i][j]  = pSrc->m_apcRefPicList[i][j];
      m_aiRefPOCList[i][j]   = pSrc->m_aiRefPOCList[i][j];
    }
  }  
  m_iDepth               = pSrc->m_iDepth;

  // referenced slice
  m_bRefenced            = pSrc->m_bRefenced;
#ifdef ROUNDING_CONTROL_BIPRED
  m_bRounding            = pSrc->m_bRounding;
#endif

  // access channel
  m_pcSPS                = pSrc->m_pcSPS;
  m_pcPPS                = pSrc->m_pcPPS;
  m_pcPic                = pSrc->m_pcPic;

  m_uiColDir             = pSrc->m_uiColDir;
  m_dLambda              = pSrc->m_dLambda;
  for (i = 0; i < 2; i++)
  {
    for (j = 0; j < MAX_NUM_REF; j++)
    {
      for (k =0; k < MAX_NUM_REF; k++)
      {
        m_abEqualRef[i][j][k] = pSrc->m_abEqualRef[i][j][k];
      }
    }
  }

  m_bNoBackPredFlag      = pSrc->m_bNoBackPredFlag;
#if MS_LCEC_LOOKUP_TABLE_EXCEPTION
  m_bRefIdxCombineCoding = pSrc->m_bRefIdxCombineCoding;
#endif
  m_uiSliceMode                   = pSrc->m_uiSliceMode;
  m_uiSliceArgument               = pSrc->m_uiSliceArgument;
  m_uiSliceCurStartCUAddr         = pSrc->m_uiSliceCurStartCUAddr;
  m_uiSliceCurEndCUAddr           = pSrc->m_uiSliceCurEndCUAddr;
  m_uiSliceIdx                    = pSrc->m_uiSliceIdx;
  m_uiEntropySliceMode            = pSrc->m_uiEntropySliceMode;
  m_uiEntropySliceArgument        = pSrc->m_uiEntropySliceArgument; 
  m_uiEntropySliceCurStartCUAddr  = pSrc->m_uiEntropySliceCurStartCUAddr;
  m_uiEntropySliceCurEndCUAddr    = pSrc->m_uiEntropySliceCurEndCUAddr;
  m_bNextSlice                    = pSrc->m_bNextSlice;
  m_bNextEntropySlice             = pSrc->m_bNextEntropySlice;
}

// ------------------------------------------------------------------------------------------------
// Sequence parameter set (SPS)
// ------------------------------------------------------------------------------------------------

TComSPS::TComSPS()
{
  // Structure
  m_uiWidth       = 352;
  m_uiHeight      = 288;
  m_uiMaxCUWidth  = 32;
  m_uiMaxCUHeight = 32;
  m_uiMaxCUDepth  = 3;
  m_uiMinTrDepth  = 0;
  m_uiMaxTrDepth  = 1;
  m_uiMaxTrSize   = 32;
  
  // Tool list
  m_bUseALF       = false;
  m_bUseDQP       = false;
  
  m_bUseMRG      = false; // SOPH:
  
  // AMVP parameter
  ::memset( m_aeAMVPMode, 0, sizeof( m_aeAMVPMode ) );
}

TComSPS::~TComSPS()
{
}

TComPPS::TComPPS()
{
#if CONSTRAINED_INTRA_PRED
  m_bConstrainedIntraPred = false;
#endif
}

TComPPS::~TComPPS()
{
}

#if SONY_QM_AVC
TComQM::TComQM()
{
  init();
#if SONY_QM_AVC_GOP_TEST
  m_iNumFrameCoded = 0;
#endif
}
TComQM::~TComQM()
{
  destroy();
}

/** set default quantization matrix to array
*/
Void TComSlice::setDefaultQM()
{
  UInt i;
  Int *puiDst=0;
  Int *puiSrc=0;

  getQM()->set_qmps_id(0);

  //4x4
  for(i=0;i<QM_LIST_NUM;i++)
  {
    puiSrc = (i<3) ? Quant_intra_default4x4 : Quant_inter_default4x4;
    puiDst = getQM()->get_QM_Org_address(getQM()->get_qmps_id(),QM4x4,i);
    ::memcpy(puiDst,puiSrc,sizeof(UInt)*16);
    puiDst = getQM()->get_QM_address(getQM()->get_qmps_id(),QM4x4,i);
    ::memcpy(puiDst,puiSrc,sizeof(UInt)*16);
  }
  //8x8
  for(i=0;i<QM_LIST_NUM;i++)
  {
    puiSrc = (i<3) ? Quant_intra_default8x8 : Quant_inter_default8x8;
    puiDst = getQM()->get_QM_Org_address(getQM()->get_qmps_id(),QM8x8,i);
    ::memcpy(puiDst,puiSrc,sizeof(UInt)*64);
    puiDst = getQM()->get_QM_address(getQM()->get_qmps_id(),QM8x8,i);
    ::memcpy(puiDst,puiSrc,sizeof(UInt)*64);
  }
  //16x16
  for(i=0;i<QM_LIST_NUM;i++)
  {
    puiSrc = (i<3) ? Quant_intra_default16x16 : Quant_inter_default16x16;
    puiDst = getQM()->get_QM_Org_address(getQM()->get_qmps_id(),QM16x16,i);
    ::memcpy(puiDst,puiSrc,sizeof(UInt)*256);
    puiDst = getQM()->get_QM_address(getQM()->get_qmps_id(),QM16x16,i);
    ::memcpy(puiDst,puiSrc,sizeof(UInt)*256);
  }
  //32x32
  for(i=0;i<QM_LIST_NUM;i++)
  {
    puiSrc = (i<3) ? Quant_intra_default32x32 : Quant_inter_default32x32;
    puiDst = getQM()->get_QM_Org_address(getQM()->get_qmps_id(),QM32x32,i);
    ::memcpy(puiDst,puiSrc,sizeof(UInt)*1024);
    puiDst = getQM()->get_QM_address(getQM()->get_qmps_id(),QM32x32,i);
    ::memcpy(puiDst,puiSrc,sizeof(UInt)*1024);
  }

}
/** check if use default quantization matrix
 * \returns true if use default quantization matrix in all size
*/
Bool TComSlice::CheckDefaultQM()
{
  UInt i;
  Int *puiDst=0;
  Int *puiSrc=0;
  UInt uiDefaultCounter=0;

  //4x4
  for(i=0;i<QM_LIST_NUM;i++)
  {
    puiSrc = (i<3) ? Quant_intra_default4x4 : Quant_inter_default4x4;
    puiDst = getQM()->get_QM_address(getQM()->get_qmps_id(),QM4x4,i);
    if(::memcmp(puiDst,puiSrc,sizeof(UInt)*16) == 0) uiDefaultCounter++;
  }

  //8x8
  for(i=0;i<QM_LIST_NUM;i++)
  {
    puiSrc = (i<3) ? Quant_intra_default8x8 : Quant_inter_default8x8;
    puiDst = getQM()->get_QM_address(getQM()->get_qmps_id(),QM8x8,i);
    if(::memcmp(puiDst,puiSrc,sizeof(UInt)*64) == 0) uiDefaultCounter++;
  }
  //16x16
  for(i=0;i<QM_LIST_NUM;i++)
  {
    puiSrc = (i<3) ? Quant_intra_default16x16 : Quant_inter_default16x16;
    puiDst = getQM()->get_QM_address(getQM()->get_qmps_id(),QM16x16,i);
    if(::memcmp(puiDst,puiSrc,sizeof(UInt)*256) == 0) uiDefaultCounter++;
  }
  //32x32
  for(i=0;i<QM_LIST_NUM;i++)
  {
    puiSrc = (i<3) ? Quant_intra_default32x32 : Quant_inter_default32x32;
    puiDst = getQM()->get_QM_address(getQM()->get_qmps_id(),QM32x32,i);
    if(::memcmp(puiDst,puiSrc,sizeof(UInt)*1024) == 0) uiDefaultCounter++;
  }
  return (uiDefaultCounter == QM_LIST_NUM * QM_SIZE_NUM) ? true : false;
}

/** get address of quantization matrix 
 * \param uiQmpsId QMPS index
 * \param uiSizeIdc side index
 * \param uiListId list index
 * \returns pointer of quantization matrix
 */
Int* TComQM::get_QM_address(UInt uiQmpsId, UInt uiSizeIdc, UInt uiListId)
{
  Int *piSrc = 0;
  switch(uiSizeIdc)
  {
    case QM4x4:
      piSrc = m_QM4x4[uiQmpsId][uiListId];
      break;
    case QM8x8:
      piSrc = m_QM8x8[uiQmpsId][uiListId];
      break;
    case QM16x16:
      piSrc = m_QM16x16[uiQmpsId][uiListId];
      break;
    case QM32x32:
      piSrc = m_QM32x32[uiQmpsId][uiListId];
      break;
    default:
      assert(0);
      piSrc = NULL;
      break;
  }
  return piSrc;
}

/** get address of default quantization matrix 
 * \param uiQmpsId QMPS index
 * \param uiSizeIdc size index
 * \param uiListId list index
 * \returns pointer of default quantization matrix
 */
Int* TComQM::get_QM_Org_address(UInt uiQmpsId, UInt uiSizeIdc, UInt uiListId)
{
  Int *piSrc = 0;
  switch(uiSizeIdc)
  {
    case QM4x4:
      piSrc = m_QM4x4_Org[uiQmpsId][uiListId];
      break;
    case QM8x8:
      piSrc = m_QM8x8_Org[uiQmpsId][uiListId];
      break;
    case QM16x16:
      piSrc = m_QM16x16_Org[uiQmpsId][uiListId];
      break;
    case QM32x32:
      piSrc = m_QM32x32_Org[uiQmpsId][uiListId];
      break;
    default:
      assert(0);
      piSrc = NULL;
      break;
  }
  return piSrc;
}
/** mode decision of quantization matrix prediction
 */
Void TComQM::xQMatrixModeDecision()
{
  Int iSizeIdc,iListIdc;
  Int *piOrg = 0;
  Int *piRecon = 0;
  Int *piResidual = new Int[1024];
  Int *piBestRecon = new Int[1024];
  UInt uiBestBit,uiQmpsId;
  UInt uiSize;
  uiQmpsId = get_qmps_id();

  for(iSizeIdc = 0; iSizeIdc < QM_SIZE_NUM; iSizeIdc++)
  {
    uiSize = (1 << (iSizeIdc+2))*(1 << (iSizeIdc+2));
    for(iListIdc = 0; iListIdc < QM_LIST_NUM; iListIdc++)
    {
      piOrg   = get_QM_Org_address(uiQmpsId, iSizeIdc, iListIdc);
      piRecon = get_QM_address(uiQmpsId, iSizeIdc, iListIdc);

      uiBestBit = MAX_UINT;

      xCalcBestBit_DPCMMode( piOrg, piRecon, piBestRecon, iSizeIdc, iListIdc, uiSize, &uiBestBit );

      ::memcpy(piRecon,piBestRecon,sizeof(Int)*uiSize);
    }
  }
  delete [] piResidual;
  delete [] piBestRecon;
}

/** change raster scan array into zigzag scan array
 * \param piSrc  raster scan array
 * \param piDist zigzag scan array
 * \param uiSizeIdc size index
 */
Void TComQM::xConvZigZag(Int *piSrc,Int *piDist, UInt uiSizeIdc)
{
  UInt i,x,y,uiSize = g_auiQMSize[uiSizeIdc];
  UInt uiSizeX   = g_auiQMSizeX[uiSizeIdc];
  UInt* pucScan  = g_auiFrameScanXY [ uiSizeIdc + 1 ];
  Int iPcmCounter = 0;

  for(i=0;i<uiSize;i++)
  {
    x = pucScan[i] % uiSizeX;
    y = pucScan[i] / uiSizeX;
      piDist[iPcmCounter]  = piSrc[pucScan[i]];
      iPcmCounter++;
  }
}
/** change zigzag scan array into raster scan array
 * \param piSrc  zigzag scan array
 * \param piDist raster scan array
 * \param uiSizeIdc size index
 */
Void TComQM::xInvZigZag(Int *piSrc,Int *piDist, UInt uiSizeIdc)
{
  UInt i,x,y,uiSize = g_auiQMSize[uiSizeIdc];
  UInt uiSizeX   = g_auiQMSizeX[uiSizeIdc];
  UInt* pucScan  = g_auiFrameScanXY [ uiSizeIdc + 1 ];
  Int iPcmCounter = 0;

  for(i=0;i<uiSize;i++)
  {
    x = pucScan[i] % uiSizeX;
    y = pucScan[i] / uiSizeX;
      piDist[pucScan[i]]  = piSrc[iPcmCounter];
      iPcmCounter++;
  }
}
/** apply DPCM to input matrix coefficient
 * \param piSrc input matrix
 * \param piDist output matrix
 * \param uiSizeIdc size index
 */
Void TComQM::xMakeDPCM(Int *piSrc,Int *piDist, UInt uiSizeIdc)
{
  UInt i,uiSize = g_auiQMSize[uiSizeIdc];
  Int  nextcoef = 0;

  for(i=0;i<uiSize;i++)
  {
    piDist[i] = piSrc[i] - nextcoef;
    nextcoef = piSrc[i];
  }
}
/** apply inverse DPCM to input matrix coefficient
 * \param piSrc  input matrix
 * \param piDist output matrix
 * \param uiSizeIdc size index
 * \param iStartValue start value of DPCM
 */
Void TComQM::xInvDPCM(Int *piSrc,Int *piDist, UInt uiSizeIdc, Int iStartValue)
{
  UInt i,uiSize = g_auiQMSize[uiSizeIdc];
  Int  lastScale = iStartValue;

  for(i=0;i<uiSize;i++)
  {
    piDist[i] = lastScale + piSrc[i];
    lastScale = piDist[i];
  }
}
/** dequantize matrix coefficient
 * \param piData input matrix coefficient
 * \param uiSizeIdc size index
 */
Void TComQM::xDequantize(Int *piData, UInt uiSizeIdc)
{
  UInt i,x,y,uiSize = g_auiQMSize[uiSizeIdc];
  UInt uiSizeX   = g_auiQMSizeX[uiSizeIdc];
  UInt* pucScan  = g_auiFrameScanXY [ uiSizeIdc + 1 ];
  Int iPcmCounter = 0;

  for(i=0;i<uiSize;i++)
  {
    x = pucScan[i] % uiSizeX;
    y = pucScan[i] / uiSizeX;
      piData[iPcmCounter] = piData[iPcmCounter];
      iPcmCounter++;
  }
}

/** calculate sum of syntax bits for matrix by applying DPCM mode and update best bit
 * \param piOrg original matrix
 * \param piRecon reconstructed matrix
 * \param piBestRecon best reconstructed matrix
 * \param iSizeIdc size index
 * \param iListIdc list index
 * \param uiSize number of matrix coefficients
 * \param uiBestBit best of sum of syntax bits for matrix
 */
Void TComQM::xCalcBestBit_DPCMMode( Int *piOrg, Int * piRecon, Int * piBestRecon, Int iSizeIdc, Int iListIdc, UInt uiSize, UInt *uiBestBit )
{
  UInt uiTargetBit=0;
  estQMStruct estQM;

  uiTargetBit = xPredDPCMQmatrix(piRecon, piOrg, iSizeIdc, &estQM);


  uiTargetBit += 4; // TableData;

  if(uiTargetBit < *uiBestBit)
  {
    *uiBestBit = uiTargetBit;
    xUpdateCondition(iSizeIdc, iListIdc, &estQM);
    ::memcpy(piBestRecon,piRecon,sizeof(Int)*uiSize);
  }
}
/** VLC or UVLC table num decision for coding residual coefficients
 * \param uiSize number of matrix coefficients
 * \param piData residual coefficients
 * \param puiTableNum best table number of VLC or UVLC
 * \returns sum of syntax bits for residual coefficients
 */
UInt TComQM::xEstResidualCode(UInt uiSize, Int *piData, UInt *puiTableNum)
{
    UInt i,uiEstBit = MAX_UINT;
    UInt uiLength = 0;
    UInt uiLengthSum;
    Bool bVlcAvailable = true;
    uiLengthSum = 0;
    for(i=0;i<uiSize;i++)
    {
      uiLength = xEstSvlc(piData[i]);
      uiLengthSum += uiLength;
      if( uiLength > 32)
      {
        bVlcAvailable = false;
      }
    }
    if(uiLengthSum < uiEstBit && bVlcAvailable == true)
    {
        uiEstBit = uiLengthSum;
    }
    return uiEstBit;
}

/** calculate bit length by using VLC
 * \param uiCodeNumber inupt code
 * \param uiTableNumber table number of VLC
 * \returns bit length
 */
UInt TComQM::xEstvlc     ( UInt uiCodeNumber, UInt uiTableNumber )
{
#if QC_BLK_CBP
  assert( uiTableNumber<=11 );
#else
  assert( uiTableNumber<=10 );
#endif
  
  UInt uiTemp;
  UInt uiLength = 0;
  UInt uiCode = 0;
  
  if (uiTableNumber == 8)
  {
      uiTableNumber = 10;
  }

  if ( uiTableNumber < 5 )
  {
    if ((Int)uiCodeNumber < (6 * (1 << uiTableNumber)))
    {
      uiTemp = 1<<uiTableNumber;
      uiCode = uiTemp+uiCodeNumber%uiTemp;
      uiLength = 1+uiTableNumber+(uiCodeNumber>>uiTableNumber);
    }
    else
    {
      uiCode = uiCodeNumber - (6 * (1 << uiTableNumber)) + (1 << uiTableNumber);
      uiLength = (6-uiTableNumber)+1+2*xLeadingZeros(uiCode);
    }
  }
  else if (uiTableNumber < 8)
  {
    uiTemp = 1<<(uiTableNumber-4);
    uiCode = uiTemp+uiCodeNumber%uiTemp;
    uiLength = 1+(uiTableNumber-4)+(uiCodeNumber>>(uiTableNumber-4));
  }
  else if (uiTableNumber == 8)
  {
    assert( uiCodeNumber<=2 );
    if (uiCodeNumber == 0)
    {
      uiCode = 1;
      uiLength = 1;
    }
    else if (uiCodeNumber == 1)
    {
      uiCode = 1;
      uiLength = 2;
    }
    else if (uiCodeNumber == 2)
    {
      uiCode = 0;
      uiLength = 2;
    }
  }
  else if (uiTableNumber == 9)
  {
    if (uiCodeNumber == 0)
    {
      uiCode = 4;
      uiLength = 3;
    }
    else if (uiCodeNumber == 1)
    {
      uiCode = 10;
      uiLength = 4;
    }
    else if (uiCodeNumber == 2)
    {
      uiCode = 11;
      uiLength = 4;
    }
    else if (uiCodeNumber < 11)
    {
      uiCode = uiCodeNumber+21;
      uiLength = 5;
    }
    else
    {
      uiTemp = 1<<4;
      uiCode = uiTemp+(uiCodeNumber+5)%uiTemp;
      uiLength = 5+((uiCodeNumber+5)>>4);
    }
  }
  else if (uiTableNumber == 10)
  {
    uiCode = uiCodeNumber+1;
    uiLength = 1+2*xLeadingZeros(uiCode);
  }
#if QC_BLK_CBP
  else if (uiTableNumber == 11)
  {
    if (uiCodeNumber == 0)
    {
      uiCode = 0;
      uiLength = 3;
    }
    else
    {
      uiCode = uiCodeNumber + 1;
      uiLength = 4;
    }
  }
#endif
  return uiLength;

}
/** count zero length
 * \param uiCode input code
 * \returns zero length
 */
UInt TComQM::xLeadingZeros(UInt uiCode)
{
  UInt uiCount = 0;
  Int iDone = 0;
  
  if (uiCode)
  {
    while (!iDone)
    {
      uiCode >>= 1;
      if (!uiCode) iDone = 1;
      else uiCount++;
    }
  }
  return uiCount;
}
/** calculate bit length by using SVLC
 * \param iCode input code
 * \returns bit length
 */
UInt TComQM::xEstSvlc     ( Int iCode )
{
  UInt uiCode;
  UInt uiLength;
  
  uiCode = xConvertToUInt( iCode );
  uiLength = xEstUvlc( uiCode );
  return uiLength;
}
/** calculate bit length by using SVLC
 * \param iCode input code
 * \param uiTableNum table number for SVLC
 * \returns bit length
 */
UInt TComQM::xEstSvlc     ( Int iCode, UInt uiTableNum )
{
  UInt uiCode;
  UInt uiLength;
  
  uiCode = xConvertToUInt( iCode );
  uiLength = xEstvlc( uiCode, uiTableNum );
  return uiLength;
}
/** calculate bit length by using UVLC
 * \param uiCode input code
 * \returns bit length
 */
UInt TComQM::xEstUvlc     ( UInt uiCode )
{
  UInt uiLength = 1;
  UInt uiTemp = ++uiCode;
  
  assert ( uiTemp );
  
  while( 1 != uiTemp )
  {
    uiTemp >>= 1;
    uiLength += 2;
  }
  
  return uiLength;
}
/** calculate bit length by using UVLC
 * \param uiCode input code
 * \param uiTableNum table number for UVLC
 * \returns bit length
 */
UInt TComQM::xEstUvlc     ( UInt uiCode, UInt uiTableNum )
{
  return xEstvlc( uiCode, uiTableNum );
}
/** apply DPCM to input matrix coefficient with quantization
 * \param piSrc input matrix
 * \param piDist output matrix
 * \param dpcm result of DPCM
 * \param uiSizeIdc size index
 */
Void TComQM::xMakeDPCMwithQ(Int* piSrc, Int* piDst, Int* dpcm, UInt uiSizeId)
{
  Int iStartValue = QM_START_VALUE;
  UInt i,x,y,uiSize = g_auiQMSize[uiSizeId];
  UInt uiSizeX = g_auiQMSizeX[uiSizeId];
  UInt* pucScan    = g_auiFrameScanXY [ uiSizeId + 1 ];
  Int  lastScale = iStartValue;
  UInt uiDataCounter = 0;
  Int sign = 0;

  for(i=0;i<uiSize;i++)
  {
    x = pucScan[i] % uiSizeX;
    y = pucScan[i] / uiSizeX;
      dpcm[uiDataCounter] = piSrc[pucScan[i]] - lastScale;
      sign = (dpcm[uiDataCounter] < 0) ? -1 : 1;
      dpcm[uiDataCounter] = abs(dpcm[uiDataCounter]) * sign;
      piDst[pucScan[i]] = lastScale + dpcm[uiDataCounter];
      lastScale = piDst[pucScan[i]];
      uiDataCounter++;
  }
}
/** estimate DPCM with Qmatrix
 * \param piDst quantized matrix
 * \param piOrg original matrix
 * \param uiSizeId size index
 * \param pestQM pointer of parameter sets of best coding mode
 * \returns sum of syntax bits for matrix
 */
UInt TComQM::xPredDPCMQmatrix(Int* piDst, Int* piOrg, UInt uiSizeId, estQMStruct *pestQM)
{
  Int dpcm[1024];
  UInt uiDataCounter = g_auiQMSize[uiSizeId];
  UInt uiEstBit = MAX_UINT;

  xMakeDPCMwithQ(piOrg, piDst, dpcm, uiSizeId);

  uiEstBit = xEstResidualCode(uiDataCounter, dpcm, &(pestQM->uiTableNumData));

  return uiEstBit;
}
/** reconstruct matrix
 * \param piRecon reconstructed matrix
 * \param piResidual residual matrix
 * \param uiSizeIdc size index
 */
Void TComQM::xCalcRecon(Int *piRecon, Int *piResidual, UInt uiSizeIdc)
{
  UInt i,uiSize = g_auiQMSize[uiSizeIdc];
  UInt* pucScan    = g_auiFrameScanXY [ uiSizeIdc + 1 ];

  for(i=0;i<uiSize;i++)
  {
    piRecon[pucScan[i]] += piResidual[pucScan[i]];
  }

}

/** parse syntax infomation 
 *  \param pchFile syntax infomation
 *  \returns true if successful
 */
Bool TComQM::xParseQM(char* pchFile)
{
  FILE *fp;
  char line[1024];
  UInt uiSizeIdc,uiListIdc;
  UInt i,uiSize,uiQmpsId = 0;
  Int *piSrc=0,data;

#if SONY_QM_AVC_GOP_TEST
  uiQmpsId = m_iNumFrameCoded%32;
#endif

  if((fp = fopen(pchFile,"r")) == (FILE*)NULL)
  {
    printf("can't open file %s :: set Default Matrix\n",pchFile);
    return true;
  }

  for(uiSizeIdc = 0; uiSizeIdc < QM_SIZE_NUM; uiSizeIdc++)
  {
    uiSize = g_auiQMSize[uiSizeIdc];
    for(uiListIdc = 0; uiListIdc < QM_LIST_NUM; uiListIdc++)
    {
      piSrc = get_QM_Org_address(uiQmpsId, uiSizeIdc, uiListIdc);

      fseek(fp,0,0);
      do 
      {
        fgets(line, 1024, fp);
        if (strstr(line, MatrixType[uiSizeIdc][uiListIdc])==NULL && feof(fp)) 
        {
          printf("Error: can't read Matrix :: set Default Matrix\n");
          return true;
        }
      }
      while (strstr(line, MatrixType[uiSizeIdc][uiListIdc]) == NULL);
      for (i=0; i<uiSize; i++)
      {
        fscanf(fp, "%d,", &data);
#if SONY_QM_AVC_GOP_TEST
        if(uiListIdc == 1) data += 2;
        if(uiListIdc == 2) data += 4;
        if(uiListIdc == 4) data += 2;
        if(uiListIdc == 5) data += 4;
        data = (Int)(16.0 + ((double)((data)-16.0) * (1.0+(double)(m_iNumFrameCoded)/16.0)));
        data = max(min(data,255),6);
#endif
        piSrc[i] = data;
      }
    }
  }
  return false;
}

/** set quantization matrix infomation
 * \param pestQM quantization matrix infomation
 * \param bIsDpcm flag for using DPCM
 * \param uiTableData VLC or UVLC Table number for data
 * \param uiTableRun VLC or UVLC Table number for run
 */
Void TComQM::set_estQM(estQMStruct *pestQM, Bool bIsDpcm, UInt uiTableData, UInt uiTableRun)
{
  pestQM->bIsDpcm = bIsDpcm;
  pestQM->uiTableNumData = uiTableData;
  pestQM->uiTableNumRun = uiTableRun;
}
/** update condition
 * \param iSizeIdc size index
 * \param iListIdc list index
 * \param pestQM quantization matrix infomation
 */
Void TComQM::xUpdateCondition(UInt iSizeIdc, UInt iListIdc, estQMStruct *pestQM)
{

  set_dpcm_flag(iSizeIdc, iListIdc, pestQM->bIsDpcm);
  set_vlc_table_num_run (iSizeIdc, iListIdc, pestQM->uiTableNumRun);
  set_vlc_table_num_data(iSizeIdc, iListIdc, pestQM->uiTableNumData);
}
/** initialization process of quantization matrix array
 */
Void TComQM::init(){
  for(UInt uiQmpsId = 0; uiQmpsId < QM_MAX_QMPS_SIZE; uiQmpsId++){
    for(UInt uiListId = 0; uiListId < QM_LIST_NUM; uiListId++){
      m_QM4x4      [uiQmpsId][uiListId] = new Int [16];
      m_QM4x4_Org  [uiQmpsId][uiListId] = new Int [16];
      m_QM8x8      [uiQmpsId][uiListId] = new Int [64];
      m_QM8x8_Org  [uiQmpsId][uiListId] = new Int [64];
      m_QM16x16    [uiQmpsId][uiListId] = new Int [256];
      m_QM16x16_Org[uiQmpsId][uiListId] = new Int [256];
      m_QM32x32    [uiQmpsId][uiListId] = new Int [1024];
      m_QM32x32_Org[uiQmpsId][uiListId] = new Int [1024];
    }
  }
}
/** destroy quantization matrix array
 */
Void TComQM::destroy(){
  for(UInt uiQmpsId = 0; uiQmpsId < QM_MAX_QMPS_SIZE; uiQmpsId++){
    for(UInt uiListId = 0; uiListId < QM_LIST_NUM; uiListId++){
      if(m_QM4x4      [uiQmpsId][uiListId]) delete [] m_QM4x4      [uiQmpsId][uiListId];
      if(m_QM4x4_Org  [uiQmpsId][uiListId]) delete [] m_QM4x4_Org  [uiQmpsId][uiListId];
      if(m_QM8x8      [uiQmpsId][uiListId]) delete [] m_QM8x8      [uiQmpsId][uiListId];
      if(m_QM8x8_Org  [uiQmpsId][uiListId]) delete [] m_QM8x8_Org  [uiQmpsId][uiListId];
      if(m_QM16x16    [uiQmpsId][uiListId]) delete [] m_QM16x16    [uiQmpsId][uiListId];
      if(m_QM16x16_Org[uiQmpsId][uiListId]) delete [] m_QM16x16_Org[uiQmpsId][uiListId];
      if(m_QM32x32    [uiQmpsId][uiListId]) delete [] m_QM32x32    [uiQmpsId][uiListId];
      if(m_QM32x32_Org[uiQmpsId][uiListId]) delete [] m_QM32x32_Org[uiQmpsId][uiListId];
    }
  }
}

#endif

